CACHEDOPS = "cachedoperations"
SLICEBUILDERS = "slicebuilders"

GENERIC = "slicebuilder"
SUBPOPULATION = "subpopulation"
ATTACK = "attack"
AUGMENTATION = "augmentation"
TRANSFORMATION = "transformation"
CURATION = "curated"
